// Uebung8 Betriebssysteme 2
// Matthias Jauernig, 2006
// Server-Programm

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <stdbool.h>
#include <signal.h>
#include <sys/wait.h>
#include <pthread.h>

#define SOCKNAME "sockunix"
#define QUEUESIZE 10

#define Fflush() while(getchar() != '\n')

bool ende = false;

void* threadfkt(void* args);
bool kommFkt(int sockd2);
void sigintfunc(int signr);

typedef struct threadParams{
	int sockd2;
	int *clicount;
	pthread_mutex_t *mutex;
} threadParams_t;


int main(void){
	int sockd, sockd2, sl, rc, scope, clicount=0;
	bool stop = false;
	struct sockaddr_un uxadr; // Bereich fuer Adressinformation
	pthread_mutex_t mutex;
	pthread_attr_t attr;
	pthread_t thrid;
	threadParams_t *args = NULL;
	
	// Mutex initialisieren
	pthread_mutex_init(&mutex,NULL);

	// Adressinformationen bereitstellen
	uxadr.sun_family = AF_UNIX;
	strcpy(uxadr.sun_path, SOCKNAME);	// File fuer Datenaustausch eintragen

	// Socket einrichten
	if((sockd = socket(AF_UNIX, SOCK_STREAM, 0)) < 0){
		perror("socket");
		return 1;
	}

	// als Vorsichtsmassnahme Socket-File loeschen - darf noch nicht existieren
	unlink(uxadr.sun_path);
	sl = sizeof(uxadr);
	if(bind(sockd, (struct sockaddr*)&uxadr, sl) < 0){ // Socketfile mit bind() anlegen
		perror("bind");
		return 2;
	}

	// Warteschlange fuer Clients anlegen
	if(listen(sockd, QUEUESIZE) < 0){
		perror("listen");
		return 3;
	}

	// Clients verarbeiten
	sigset(SIGINT, sigintfunc);	// SIGINT verarbeiten unter Unix
	printf("Warte auf Client-Anforderungen...\n");
	pthread_attr_init(&attr);	// Thread-Attribute initialisieren
	scope = PTHREAD_SCOPE_SYSTEM;	// Threads in Konkurrenz um CPU mit allen Prozessen
	if((rc=pthread_attr_setscope(&attr,scope)) != 0){
		fprintf(stderr,"pthread_attr_setscope: %s\n", strerror(rc));
		return 4;
	}
	do{
		// auf Client-Anforderungen warten
		if((sockd2 = accept(sockd, 0, 0)) < 0 && !ende){
			perror("accept");
			return 5;
		}

		if(ende){
			break;
		}

		args = (threadParams_t*)malloc(sizeof(threadParams_t));
		args->clicount = &clicount;
		args->mutex = &mutex;
		args->sockd2 = sockd2;
		
		// neuen Thread starten
		if((rc=pthread_create(&thrid,&attr,threadfkt,args)) != 0){
			fprintf(stderr,"pthread_create: %s\n", strerror(rc));
			return 6;
		}
		
	}while(!ende); // Ende ueber SIGINT

	// sockd schliessen
	if(close(sockd) < 0){
		perror("close(sockd)");
		return 7;
	}

	// auf Ende aller Threads warten
	printf("Warte auf Thread-Beendigungen...\n");
	while(!stop){
		pthread_mutex_lock(&mutex);
		stop = (clicount<=0);
		pthread_mutex_unlock(&mutex);
	}

	// Socket entfernen
	if(unlink(uxadr.sun_path) < 0){
		perror("unlink");
		return 8;
	}

	pthread_mutex_destroy(&mutex);
	return 0;
}


void sigintfunc(int signr){
	switch(signr){
	case SIGINT:
		printf("SIGINT empfangen...\n");
		ende = true;
	}
}

void* threadfkt(void* args){
	threadParams_t* largs = (threadParams_t*)args;
	
	pthread_mutex_lock(largs->mutex);
	(*(largs->clicount))++;
	printf("%d: Neuer Client angemeldet, Clients: %d\n", pthread_self(), *(largs->clicount));
	pthread_mutex_unlock(largs->mutex);	

	if(!kommFkt(largs->sockd2)){
		printf("%d: Fehler aufgetreten...\n", pthread_self());
	}

	pthread_mutex_lock(largs->mutex);
	(*(largs->clicount))--;
	printf("%d: Client abgemeldet, Clients: %d\n", pthread_self(), *(largs->clicount));
	pthread_mutex_unlock(largs->mutex);
	
	free(largs);
	
	return NULL;
}

bool kommFkt(int sockd2){
	int l;
	char eab[100]; // E/A-Bereich
	FILE *fp;
	char fname[100]; // Dateiname
	
	// Kommunikation ueber sockd2
	// 1.) Dateiname empfangen und Datei zum Schreiben oeffnen
	if((l = recv(sockd2, eab, 100, 0)) < 0){
		perror("recv");
	}
	strcpy(fname, eab);
	if((fp = fopen(eab, "w")) < 0){
		perror("fopen");
		return false;
	}
	printf("%d: Schreibe Daten von Client in Datei '%s'\n", pthread_self(), eab);
	
	// 2.) Daten empfangen und in Datei schreiben
	do{
		if((l = recv(sockd2, eab, 100, 0)) < 0){
			perror("recv");
		}
		if(l>0)
		{
			printf("%d: Daten von Client empfangen, %d Bytes: %s", pthread_self(), l, eab);
			// Empfangene Daten in Datei schreiben
			if(fwrite(&eab, strlen(eab)+1, 1, fp) < 0){
				perror("fwrite");
				return false;
			}
			// Byteanzahl als Quittung zuruecksenden
			sprintf(eab, "%d", l);
			if((l=send(sockd2, eab, strlen(eab)+1, 0)) < 0){
				perror("send");
			}
		}
	}while(l != 0);
	
	fclose(fp);
	if(close(sockd2) < 0){
		perror("close(sockd2)");
		return false;
	}
	
	return true;
}
